//======================================================================
//  flaa; main
//    Author: Dave Freese, W1HKJ <w1hkj@w1hkj.com>
//
//  This program is distributed under the GPL license
//======================================================================

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <fstream>
#include <ostream>
#include <cstring>
#include <strstream>
#include <ctime>
#include <vector>

#include <FL/Fl.H>
#include <FL/Enumerations.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Sys_Menu_Bar.H>
#include <FL/x.H>
#include <FL/Fl_Help_Dialog.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/Fl_Browser.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Color_Chooser.H>

#include <config.h>

#include "flaa.h"
#include "flaa_ui.h"

#include "status.h"
#include "util.h"
#include "data_io.h"
#include "fileselect.h"

#ifdef WIN32
#  include "flaarc.h"
#  include "compat.h"
#  define dirent fl_dirent_no_thanks
#endif

#include <FL/filename.H>

#ifdef __MINGW32__
#	if FLAA_FLTK_API_MAJOR == 1 && FLAA_FLTK_API_MINOR < 3
#		undef dirent
#		include <dirent.h>
#	endif
#	ifndef IDI_ICON
#		define IDI_ICON 101
#	endif
#else
#	include <dirent.h>
#	include "flaa_icon.cxx"
#endif

#include <FL/x.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Image.H>
#include <FL/Fl_File_Chooser.H>

Fl_Double_Window *main_window = (Fl_Double_Window *)0;

Cserial aa_ser_port;

char serbuff[BUFFSIZE];

bool is_connected = false;

vector<FRX> testdata;

string flaa_home_dir;

static void checkdirectories(string appdir)
{
	char dirbuf[FL_PATH_MAX + 1];
	size_t p;
#ifdef __WIN32__
	p = appdir.rfind("flaa.exe");
	if (p != string::npos) appdir.erase(p);
	p = appdir.find("FL_APPS\\");
	if (p != string::npos) {
		flaa_home_dir.assign(appdir.substr(0, p + 8));
		flaa_home_dir.append("flaa.files\\");
	} else if (flaa_home_dir.empty()) {
		fl_filename_expand(dirbuf, sizeof(dirbuf) -1, "$USERPROFILE/");
		flaa_home_dir.assign(dirbuf);
		flaa_home_dir.append("flrig.files\\");
	}
#else
	p = appdir.rfind("flaa");
	if (p != std::string::npos)
		appdir.erase(p);
	p = appdir.find("FL_APPS/");
	if (p != string::npos)
		flaa_home_dir.assign(appdir.substr(0, p + 8));
	if (flaa_home_dir.empty()) {
		fl_filename_expand(dirbuf, FL_PATH_MAX, "$HOME/");
	flaa_home_dir.assign(dirbuf);
	}
	flaa_home_dir.append(".flaa/");
#endif

	struct {
		string& dir;
		const char* suffix;
		void (*new_dir_func)(void);
	} dirs[] = {
		{ flaa_home_dir, 0, 0 }
	};

	int r;
	for (size_t i = 0; i < sizeof(dirs)/sizeof(*dirs); i++) {
		if (dirs[i].suffix)
			dirs[i].dir.assign(flaa_home_dir).append(dirs[i].suffix).append("/");

		if ((r = mkdir(dirs[i].dir.c_str(), 0777)) == -1 && errno != EEXIST) {
			cerr << "Could not make directory" << ' ' << dirs[i].dir
				 << ": " << strerror(errno) << '\n';
			exit(EXIT_FAILURE);
		}
		else if (r == 0 && dirs[i].new_dir_func)
			dirs[i].new_dir_func();
	}
}

string aa_model;

string hexstr(string s)
{
	static string s2;
	s2.clear();
	for (size_t n = 0; n < s.length(); n++)
		if (s[n] == 0x0D) s2.append("<CR>");
		else if (s[n] == 0x0A) s2.append("<LF>\n");
		else s2 += s[n];
	return s2;
}

#if !defined(__WIN32__) && !defined(__APPLE__)
Pixmap	flaa_icon_pixmap;

void make_pixmap(Pixmap *xpm, const char **data)
{
	Fl_Window w(0,0, PACKAGE_NAME);
	w.xclass(PACKAGE_NAME);
	w.show();
	w.make_current();
	Fl_Pixmap icon(data);
	int maxd = (icon.w() > icon.h()) ? icon.w() : icon.h();
	*xpm = fl_create_offscreen(maxd, maxd);
	fl_begin_offscreen(*xpm);
	fl_color(FL_BACKGROUND_COLOR);
	fl_rectf(0, 0, maxd, maxd);
	icon.draw(maxd - icon.w(), maxd - icon.h());
	fl_end_offscreen();
}

#endif

int parse_args(int argc, char **argv, int& idx);

char title[50];
char progdir[80];

int handle (int e) {
	if (Fl::focus() == main_window)
		return (e == FL_SHORTCUT);
	return 0;
}

void visit_URL(void* arg)
{
	const char* url = reinterpret_cast<const char *>(arg);
#ifndef __WOE32__
	const char* browsers[] = {
#  ifdef __APPLE__
		getenv("FLDIGI_BROWSER"), // valid for any OS - set by user
		"open"                    // OS X
#  else
		"fl-xdg-open",            // Puppy Linux
		"xdg-open",               // other Unix-Linux distros
		getenv("FLDIGI_BROWSER"), // force use of spec'd browser
		getenv("BROWSER"),        // most Linux distributions
		"sensible-browser",
		"firefox",
		"mozilla"                 // must be something out there!
#  endif
	};
	switch (fork()) {
	case 0:
#  ifndef NDEBUG
		unsetenv("MALLOC_CHECK_");
		unsetenv("MALLOC_PERTURB_");
#  endif
		for (size_t i = 0; i < sizeof(browsers)/sizeof(browsers[0]); i++)
			if (browsers[i])
				execlp(browsers[i], browsers[i], url, (char*)0);
		exit(EXIT_FAILURE);
	case -1:
		fl_alert(
"Could not run a web browser:\n%s\n\n"
"Open this URL manually:\n%s",
			 strerror(errno), url);
	}
#else
	if ((int)ShellExecute(NULL, "open", url, NULL, NULL, SW_SHOWNORMAL) <= 32)
		fl_alert("Could not open url:\n%s\n", url);
#endif
}

void exit_main(Fl_Widget *w)
{
	if (Fl::event_key() == FL_Escape)
		return;
	flaa_status.saveLastState();
	aa_ser_port.ClosePort();
	main_window->hide();
}

void cb_exit()
{
	flaa_status.saveLastState();
	aa_ser_port.ClosePort();
	main_window->hide();
	exit_read_thread();
}

void init_vals()
{
	static char sz_value[50];
	inp_ser_device->value(flaa_status.serial_port.c_str());

	snprintf(sz_value, sizeof(sz_value), "%6.4f", flaa_status.fmin);
	inp_min_freq->value(sz_value);

	snprintf(sz_value, sizeof(sz_value), "%6.4f", flaa_status.fmax);
	inp_max_freq->value(sz_value);

	snprintf(sz_value, sizeof(sz_value), "%d", flaa_status.npts);
	inp_nr_pts->value(sz_value);

	snprintf(sz_value, sizeof(sz_value), "%4.1f", flaa_status.system_z);
	inp_system_z->value(sz_value);

	freq_80->value(flaa_status.freq_80);
	freq_60->value(flaa_status.freq_60);
	freq_40->value(flaa_status.freq_40);
	freq_30->value(flaa_status.freq_30);
	freq_20->value(flaa_status.freq_20);
	freq_15->value(flaa_status.freq_15);
	freq_10->value(flaa_status.freq_10);
	freq_6->value(flaa_status.freq_6);
	freq_2->value(flaa_status.freq_2);
	freq_220->value(flaa_status.freq_220);
	freq_440->value(flaa_status.freq_440);
	freq_other->value(flaa_status.freq_other);

	prg_swr_80->value();  prg_swr_80->label("");  prg_swr_80->redraw_label();
	prg_swr_60->value();  prg_swr_60->label("");  prg_swr_60->redraw_label();
	prg_swr_40->value();  prg_swr_40->label("");  prg_swr_40->redraw_label();
	prg_swr_30->value();  prg_swr_30->label("");  prg_swr_30->redraw_label();
	prg_swr_20->value();  prg_swr_20->label("");  prg_swr_20->redraw_label();
	prg_swr_15->value();  prg_swr_15->label("");  prg_swr_15->redraw_label();
	prg_swr_10->value();  prg_swr_10->label("");  prg_swr_10->redraw_label();
	prg_swr_6->value();   prg_swr_6->label("");   prg_swr_6->redraw_label();
	prg_swr_2->value();   prg_swr_2->label("");   prg_swr_2->redraw_label();
	prg_swr_220->value(); prg_swr_220->label(""); prg_swr_220->redraw_label();
	prg_swr_440->value(); prg_swr_440->label(""); prg_swr_440->redraw_label();
	prg_swr_other->value(); prg_swr_other->label(""); prg_swr_other->redraw_label();

	inp_carrier_freq->value(flaa_status.carrier_freq);
	inp_cable_freq->value(flaa_status.cable_freq);
	inp_velocity_factor->value(flaa_status.velocity_factor);
	btn_units->value(flaa_status.units);

	inp_frequency_correction->value(flaa_status.freq_corr);

static int col_widths[] = {78 * brws_vf->w() / 100, 0};
	brws_vf->column_widths(col_widths);
	brws_vf->column_char('\t');
	brws_vf->add("RG-6/U PE (Belden 8215)\t0.66");
	brws_vf->add("RG-6/U Foam (Belden 9290)\t0.91");
	brws_vf->add("RG-8/U (PE (Belden 8237)\t0.66");
	brws_vf->add("RG-8/U Foam (Belden 8214)\t0.78");
	brws_vf->add("RG-8/U (Belden 9913)\t0.84");
	brws_vf->add("RG-8X (Belden 9258)\t0.82");
	brws_vf->add("RG-11/U Foam HDPE (Beld. 9292)\t0.84");
	brws_vf->add("RG-58/U PE (Belden 9201)\t0.66");
	brws_vf->add("RG-58A/U Foam (Belden 8219)\t0.73");
	brws_vf->add("RG-59A/U PE (Belden 8241)\t0.66");
	brws_vf->add("RG-59A/U Foam (Belden 8241F)\t0.78");
	brws_vf->add("RG-174 PE (Belden 8216)\t0.66");
	brws_vf->add("RG-174 Foam (Belden 7805R)\t0.735");
	brws_vf->add("RG-213/U (Belden 8267)\t0.66");
	brws_vf->add("LMR-240\t0.84");
	brws_vf->add("LMR-240UF\t0.84");
	brws_vf->add("LMR-400\t0.85");
	brws_vf->add("LMR-400UF\t0.85");
	brws_vf->add("Davis BuryFlex\t0.83");

}

void vf_select()
{
	int row = brws_vf->value();
	string s = brws_vf->text(row);
	size_t p = s.find("\t");
	s.erase(0, p+1);
	float vf = 0.66;
	sscanf(s.c_str(), "%f", &vf);
	flaa_status.velocity_factor = vf;
	inp_velocity_factor->value(flaa_status.velocity_factor);
	update_cable_values(0);
}

int main(int argc, char **argv)
{
	int arg_idx;

	checkdirectories(argv[0]);

	Fl::args(argc, argv, arg_idx, parse_args);

	Fl::add_handler (handle);

	Fl::visual (FL_DOUBLE|FL_INDEX);

	Fl::lock();

	main_window = make_flaa_win();

	flaa_status.loadLastState();

	size_t p = 0;
	string ports = flaa_status.port_ids;
	while((p = ports.find("\n")) != string::npos) {
		inp_ser_device->add(ports.substr(0, p).c_str());
		ports.erase(0, p+1);
	}
	init_vals();

	btn_connect->label("Connect");
	btn_connect->redraw_label();
	if (btn_toggle_status->value()) {
		txt_status->value("Not connected");
		txt_status->redraw();
}
	config_axis();

	sprintf (title, "flaa %s", FLAA_VERSION);
	main_window->label(title);
	main_window->callback(exit_main);

	if (!start_read_thread()) return 1;

#ifdef WIN32
	main_window->icon((char*)LoadIcon(fl_display, MAKEINTRESOURCE(IDI_ICON)));
	main_window->show(argc, argv);
#elif !defined(__APPLE__)
	make_pixmap(&flaa_icon_pixmap, flaa_xpm);
	main_window->icon((char *)flaa_icon_pixmap);
	main_window->show(argc, argv);
#else
	main_window->show(argc, argv);
#endif

  while (Fl::wait() > 0);

  return 0;
}

int parse_args(int argc, char **argv, int& idx)
{
	if (strcasecmp("--help", argv[idx]) == 0) {
		printf("Usage: \n\
  --help this help text\n\
  --version\n");
		exit(0);
	}
	if (strcasecmp("--version", argv[idx]) == 0) {
		printf("Version: " VERSION "\n");
		exit (0);
	}
	return 0;
}

void cb_ser_device()
{
}

void cb_min_freq()
{
	float minf;
	sscanf(inp_min_freq->value(), "%f", &minf);
	flaa_status.fmin = minf;
	swr_canvas->clear();
	rx_canvas->clear();
	config_axis();
}

void cb_max_freq()
{
	float maxf;
	sscanf(inp_max_freq->value(), "%f", &maxf);
	flaa_status.fmax = maxf;
	swr_canvas->clear();
	rx_canvas->clear();
	config_axis();
}

void cb_nr_pts()
{
	int npts;
	sscanf(inp_nr_pts->value(), "%d", &npts);
	flaa_status.npts = npts;
	swr_canvas->clear();
	rx_canvas->clear();
}

void cb_Z0()
{
	float z;
	sscanf(inp_system_z->value(), "%f", &z);
	if (z <= 0.1) {
		z = 50;
		inp_system_z->value("50");
	}
	flaa_status.system_z = z;
	swr_canvas->clear();
	rx_canvas->clear();
}

void cb_run()
{	if (!is_connected) {
		if (btn_toggle_status->value()) {
			txt_status->value("Not Connected!");
			txt_status->redraw();
		}
		return;
	}
	if (!rx_data) rx_data = true;
}

int baud[] = { 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800 };

void cb_connect()
{
	if (is_connected) {
		aa_ser_port.ClosePort();
		btn_connect->label("Connect");
		btn_connect->redraw_label();
		bx_connected->color(FL_BACKGROUND_COLOR);
		bx_connected->redraw();
		is_connected = false;
		if (btn_toggle_status->value()) {
			txt_status->value("Not connected");
			txt_status->redraw();
		}
		return;
	}
	aa_ser_port.Baud(baud[flaa_status.baudrate]);
	aa_ser_port.Stopbits(1);
	aa_ser_port.Device(flaa_status.serial_port);
	if (aa_ser_port.OpenPort()) {
		aa_ser_port.WriteBuffer("VER\n", 4);
		int nread = 0;
		int numtries = 100;
		while ((nread = aa_ser_port.ReadBuffer(serbuff, 128)) == 0) {
			MilliSleep(10);
			if (!numtries--) {
				btn_connect->label("Connect");
				btn_connect->redraw_label();
				bx_connected->color(FL_BACKGROUND_COLOR);
				bx_connected->redraw();
				is_connected = false;
				return;
			}
		}
		string sbuff = serbuff;
		if (sbuff.find("OK") != string::npos) {
			aa_model = sbuff;
			size_t p = aa_model.find("AA-");
			aa_model.erase(0,p);
			p = aa_model.find(" ");
			aa_model.erase(p);
			btn_connect->label(aa_model.c_str());
			btn_connect->redraw_label();
			bx_connected->color(FL_GREEN);
			bx_connected->redraw();
			if (btn_toggle_status->value()) {
				txt_status->value("Connected");
				txt_status->redraw();
			}
			is_connected = true;
		} else {
			is_connected = false;
			btn_connect->label("Connect");
			btn_connect->redraw_label();
			bx_connected->color(FL_BACKGROUND_COLOR);
			bx_connected->redraw();
			if (btn_toggle_status->value()) {
				txt_status->value("Not connected");
				txt_status->redraw();
			}
		}
	}
}

void xaxis_config(
		Ca_X_Axis *axis,
		int lblsize,
		int scaletype,
		double minval, double maxval,
		double major_step, double label_step,
		const char *format,
		Fl_Color minor_color, Fl_Color major_color, Fl_Color lbl_color,
		int grid_visibility,
		Fl_Color axis_color,
		int align ) {
	axis->labelsize(lblsize);
	axis->scale(scaletype);
	axis->minimum(minval);
	axis->maximum(maxval);
	axis->major_step(major_step);
	axis->label_step(label_step);
	axis->label_format(format);
	axis->minor_grid_color(minor_color);
	axis->major_grid_color(major_color);
	axis->label_grid_color(lbl_color);
	axis->grid_visible(grid_visibility);
	axis->axis_color(axis_color);
	axis->axis_align(align);
}

void yaxis_config(
		Ca_Y_Axis *axis,
		int lblsize,
		int scaletype,
		double minval, double maxval,
		double major_step, double label_step,
		const char *format,
		Fl_Color minor_color, Fl_Color major_color, Fl_Color lbl_color,
		int grid_visibility,
		Fl_Color axis_color,
		int align ) {
	axis->labelsize(lblsize);
	axis->scale(scaletype);
	axis->minimum(minval);
	axis->maximum(maxval);
	axis->major_step(major_step);
	axis->label_step(label_step);
	axis->label_format(format);
	axis->minor_grid_color(minor_color);
	axis->major_grid_color(major_color);
	axis->label_grid_color(lbl_color);
	axis->grid_visible(grid_visibility);
	axis->axis_color(axis_color);
	axis->axis_align(align);
}

const char *fformat = "%g";
void config_axis()
{
	xaxis_config(
		swr_xaxis, 9, CA_LIN,
		flaa_status.fmin, flaa_status.fmax,
		0.05, 0.05, fformat,
		fl_gray_ramp(20), fl_gray_ramp(15), fl_gray_ramp(10),
		CA_MINOR_GRID | CA_MAJOR_GRID | CA_LABEL_GRID,
		FL_BLACK,
		CA_BOTTOM | CA_LINE);
	swr_xaxis->redraw();

	yaxis_config(
		swr_yaxis, 9, CA_LIN, 1.0, 6.0, 1.0, 1.0, fformat,
		fl_gray_ramp(20), fl_gray_ramp(15), fl_gray_ramp(10),
		CA_MINOR_GRID | CA_MAJOR_GRID | CA_LABEL_GRID,
		FL_BLACK,
		CA_LEFT );
	swr_yaxis->redraw();

	xaxis_config(
		rx_freq_xaxis, 9, CA_LIN,
		flaa_status.fmin, flaa_status.fmax,
		0.05, 0.05, fformat,
		fl_gray_ramp(20), fl_gray_ramp(15), fl_gray_ramp(10),
		CA_MINOR_GRID | CA_MAJOR_GRID | CA_LABEL_GRID,
		FL_BLACK,
		CA_BOTTOM | CA_LINE);
	rx_freq_xaxis->redraw();

	yaxis_config(
		R_yaxis, 9, CA_LIN, 0.0, 100.0, 10, 10, fformat,
		fl_gray_ramp(20), fl_gray_ramp(15), fl_gray_ramp(10),
		CA_MINOR_GRID | CA_MAJOR_GRID | CA_LABEL_GRID,
		FL_BLACK,
		CA_LEFT );
	R_yaxis->redraw();

	yaxis_config(
		X_yaxis, 9, CA_LIN, -100.0, 100.0, 10, 10, fformat,
		fl_gray_ramp(20), fl_gray_ramp(15), fl_gray_ramp(10),
		CA_MINOR_GRID | CA_MAJOR_GRID | CA_LABEL_GRID,
		FL_BLACK,
		CA_RIGHT );
	X_yaxis->redraw();

	main_window->redraw();
}

void  export_data()
{
	if (testdata.empty()) return;

	const char *fn = FSEL::saveas("Export data", "CSV file\t*.txt");
	if (fn) {
		ofstream csvfile(fn, ios::binary);
		for (size_t n = 0; n < testdata.size(); n++) {
			csvfile << testdata[n].f << "," << testdata[n].r << "," << testdata[n].x << std::endl;
		}
		csvfile.close();
	}
}

void import_data()
{
	const char *fn = FSEL::select("Import data", "CSV file\t*.txt");
	if (fn) {
		double f, r, x;
		char c;
		ifstream csvfile(fn, ios::binary);
		testdata.clear();
		while (!csvfile.eof()) {
			csvfile >> f >> c >> r >> c >> x;
			testdata.push_back(FRX(f, r, x));
		}
		csvfile.close();
	}
	size_t npts = testdata.size();
	if (!npts) return;
	flaa_status.fmin = testdata[0].f;
	flaa_status.fmax = testdata[npts-1].f;
	flaa_status.npts = npts;

	char value[20];
	snprintf(value, sizeof(value), "%d", flaa_status.npts);
	inp_nr_pts->value(value);
	snprintf(value, sizeof(value), "%f", flaa_status.fmin);
	inp_min_freq->value(value);
	snprintf(value, sizeof(value), "%f", flaa_status.fmax);
	inp_max_freq->value(value);

	config_axis();

	plot_data(NULL);
}

void  plot_data(void *)
{

	if (testdata.empty()) {
#if DEBUG
std::cout << "testdata empty()\n";
#endif
		rx_canvas->clear();
		swr_canvas->clear();
		return;
	}

#if DEBUG
std::cout << "========================================\n";
for (size_t n = 0; n < testdata.size(); n++)
std::cout << "n: " << n << ", F: " << testdata[n].f << ", R: " << testdata[n].r << ", X: " << testdata[n].x << std::endl;
std::cout << "========================================\n";
#endif

	Ca_LinePoint *P_R = 0;
	Ca_LinePoint *P_X = 0;
	Ca_LinePoint *P_SWR = 0;
	FRX data;

double maxR = 1, maxX = 1;
double d;

	rx_canvas->clear();
	Ca_Canvas::current(rx_canvas);
	rx_canvas->current_x(rx_freq_xaxis);
	rx_canvas->current_y(R_yaxis);

	R_yaxis->minimum(0);
	R_yaxis->maximum(maxR);
	for (size_t n = 0; n < testdata.size(); n++) {
		d = testdata[n].r;
		P_R = new Ca_LinePoint(
				P_R, testdata[n].f ,
				d,
				2,
				flaa_status.r_color,
				flaa_status.r_point);
		if (d > maxR)
			maxR = 10 * (d / 10 + 1);
	}
	R_yaxis->maximum(maxR);
	R_yaxis->redraw();

	rx_canvas->current_y(X_yaxis);
	X_yaxis->minimum(-maxX);
	X_yaxis->maximum(maxX);

	for (size_t n = 0; n < testdata.size(); n++) {
		d = testdata[n].x;
		P_X = new Ca_LinePoint(
				P_X, testdata[n].f,
				d,
				2,
				flaa_status.x_color,
				flaa_status.x_point);
		if (fabs(d) > maxX)
			maxX = 10 * ((int)fabs(d) / 10 + 1);
		if (d > maxX)
			maxX = 10 * ((int)d / 10 + 1);
	}
	X_yaxis->minimum(-maxX);
	X_yaxis->maximum(maxX);
	X_yaxis->redraw();
	rx_canvas->redraw();

	swr_canvas->clear();
	Ca_Canvas::current(swr_canvas);
	swr_canvas->current_x(swr_xaxis);
	swr_canvas->current_y(swr_yaxis);

	double maxSWR = 5;
	swr_yaxis->maximum(maxSWR);

	double R, X, B, SWR;
	float Z0 = flaa_status.system_z;
	for (size_t n = 0; n < testdata.size(); n++) {
		R = testdata[n].r / Z0;
		X = testdata[n].x / Z0;
		if (R == 0) SWR = 1;
		B = (X * X + 1) / R + R;
		SWR = (B + sqrtf( B * B - 4)) / 2.0;
		P_SWR = new Ca_LinePoint(
					P_SWR, testdata[n].f,
					SWR,
					2,
					flaa_status.swr_color,
					flaa_status.swr_point);
		if (SWR > maxSWR)
			maxSWR = 5 * ((int)SWR / 5 + 1);
	}
	swr_yaxis->maximum(maxSWR);
	swr_yaxis->redraw();
	swr_canvas->redraw();
}

void swr_stop_all()
{
	flaa_status.swr_80_on = false;  btn_run_80->value(0);
	flaa_status.swr_60_on = false;  btn_run_60->value(0);
	flaa_status.swr_40_on = false;  btn_run_40->value(0);
	flaa_status.swr_30_on = false;  btn_run_30->value(0);
	flaa_status.swr_20_on = false;  btn_run_20->value(0);
	flaa_status.swr_15_on = false;  btn_run_15->value(0);
	flaa_status.swr_10_on = false;  btn_run_10->value(0);
	flaa_status.swr_6_on = false;   btn_run_6->value(0);
	flaa_status.swr_2_on = false;   btn_run_2->value(0);
	flaa_status.swr_220_on = false; btn_run_220->value(0);
	flaa_status.swr_440_on = false; btn_run_440->value(0);
	flaa_status.swr_other_on = false;   btn_run_other->value(0);
}

void clear_swrs()
{
	prg_swr_80->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_80->label(""); prg_swr_80->redraw_label();

	prg_swr_60->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_60->label(""); prg_swr_60->redraw_label();

	prg_swr_40->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_40->label(""); prg_swr_40->redraw_label();

	prg_swr_30->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_30->label(""); prg_swr_30->redraw_label();

	prg_swr_20->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_20->label(""); prg_swr_20->redraw_label();

	prg_swr_15->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_15->label(""); prg_swr_15->redraw_label();

	prg_swr_10->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_10->label(""); prg_swr_10->redraw_label();

	prg_swr_6->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_6->label(""); prg_swr_6->redraw_label();

	prg_swr_2->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_2->label(""); prg_swr_2->redraw_label();

	prg_swr_220->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_220->label(""); prg_swr_2->redraw_label();

	prg_swr_440->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_440->label(""); prg_swr_440->redraw_label();

	prg_swr_other->selection_color(FL_BACKGROUND2_COLOR);
	prg_swr_other->label(""); prg_swr_2->redraw_label();

}

double swr_on_80;
void update_swr_80(void *)
{
	prg_swr_80->value(100 * (swr_on_80 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_80);
	prg_swr_80->label(label);
	prg_swr_80->redraw_label();
	if (swr_on_80 < 2.0) prg_swr_80->selection_color(FL_GREEN);
	else if (swr_on_80 < 3.0) prg_swr_80->selection_color(FL_YELLOW);
	else prg_swr_80->selection_color(FL_RED);
}

double swr_on_60;
void update_swr_60(void *)
{
	prg_swr_60->value(100 * (swr_on_60 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_60);
	prg_swr_60->label(label);
	prg_swr_60->redraw_label();
	if (swr_on_60 < 2.0) prg_swr_60->selection_color(FL_GREEN);
	else if (swr_on_60 < 3.0) prg_swr_60->selection_color(FL_YELLOW);
	else prg_swr_60->selection_color(FL_RED);
}

double swr_on_40;
void update_swr_40(void *)
{
	prg_swr_40->value(100 * (swr_on_40 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_40);
	prg_swr_40->label(label);
	prg_swr_40->redraw_label();
	if (swr_on_40 < 2.0) prg_swr_40->selection_color(FL_GREEN);
	else if (swr_on_40 < 3.0) prg_swr_40->selection_color(FL_YELLOW);
	else prg_swr_40->selection_color(FL_RED);
}

double swr_on_30;
void update_swr_30(void *)
{
	prg_swr_30->value(100 * (swr_on_30 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_30);
	prg_swr_30->label(label);
	prg_swr_30->redraw_label();
	if (swr_on_30 < 2.0) prg_swr_30->selection_color(FL_GREEN);
	else if (swr_on_30 < 3.0) prg_swr_30->selection_color(FL_YELLOW);
	else prg_swr_30->selection_color(FL_RED);
}

double swr_on_20;
void update_swr_20(void *)
{
	prg_swr_20->value(100 * (swr_on_20 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_20);
	prg_swr_20->label(label);
	prg_swr_20->redraw_label();
	if (swr_on_20 < 2.0) prg_swr_20->selection_color(FL_GREEN);
	else if (swr_on_20 < 3.0) prg_swr_20->selection_color(FL_YELLOW);
	else prg_swr_20->selection_color(FL_RED);
}

double swr_on_15;
void update_swr_15(void *)
{
	prg_swr_15->value(100 * (swr_on_15 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_15);
	prg_swr_15->label(label);
	prg_swr_15->redraw_label();
	if (swr_on_15 < 2.0) prg_swr_15->selection_color(FL_GREEN);
	else if (swr_on_15 < 3.0) prg_swr_15->selection_color(FL_YELLOW);
	else prg_swr_15->selection_color(FL_RED);
}

double swr_on_10;
void update_swr_10(void *)
{
	prg_swr_10->value(100 * (swr_on_10 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_10);
	prg_swr_10->label(label);
	prg_swr_10->redraw_label();
	if (swr_on_10 < 2.0) prg_swr_10->selection_color(FL_GREEN);
	else if (swr_on_10 < 3.0) prg_swr_10->selection_color(FL_YELLOW);
	else prg_swr_10->selection_color(FL_RED);
}

double swr_on_6;
void update_swr_6(void *)
{
	prg_swr_6->value(100 * (swr_on_6 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_6);
	prg_swr_6->label(label);
	prg_swr_6->redraw_label();
	if (swr_on_6 < 2.0) prg_swr_6->selection_color(FL_GREEN);
	else if (swr_on_6 < 3.0) prg_swr_6->selection_color(FL_YELLOW);
	else prg_swr_6->selection_color(FL_RED);
}

double swr_on_2;
void update_swr_2(void *)
{
	prg_swr_2->value(100 * (swr_on_2 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_2);
	prg_swr_2->label(label);
	prg_swr_2->redraw_label();
	if (swr_on_2 < 2.0) prg_swr_2->selection_color(FL_GREEN);
	else if (swr_on_2 < 3.0) prg_swr_2->selection_color(FL_YELLOW);
	else prg_swr_2->selection_color(FL_RED);
}

double swr_on_220;
void update_swr_220(void *)
{
	prg_swr_220->value(100 * (swr_on_440 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_220);
	prg_swr_220->label(label);
	prg_swr_220->redraw_label();
	if (swr_on_220 < 2.0) prg_swr_220->selection_color(FL_GREEN);
	else if (swr_on_220 < 3.0) prg_swr_220->selection_color(FL_YELLOW);
	else prg_swr_220->selection_color(FL_RED);
}

double swr_on_440;
void update_swr_440(void *)
{
	prg_swr_440->value(100 * (swr_on_440 - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_440);
	prg_swr_440->label(label);
	prg_swr_440->redraw_label();
	if (swr_on_440 < 2.0) prg_swr_440->selection_color(FL_GREEN);
	else if (swr_on_440 < 3.0) prg_swr_440->selection_color(FL_YELLOW);
	else prg_swr_440->selection_color(FL_RED);
}

double swr_on_other;
void update_swr_other(void *)
{
	prg_swr_other->value(100 * (swr_on_other - 1.0) / 8.0);
	static char label[20];
	snprintf(label, sizeof(label), "%.1f", swr_on_other);
	prg_swr_other->label(label);
	prg_swr_other->redraw_label();
	if (swr_on_other < 2.0) prg_swr_other->selection_color(FL_GREEN);
	else if (swr_on_other < 3.0) prg_swr_other->selection_color(FL_YELLOW);
	else prg_swr_other->selection_color(FL_RED);
}

void run_80()
{
	flaa_status.swr_80_on = btn_run_80->value();
	if (flaa_status.swr_80_on) {
		swr_on_80 = 2.5;
		update_swr_80(NULL);
	}
}

void run_60()
{
	flaa_status.swr_60_on = btn_run_60->value();
	if (flaa_status.swr_60_on) {
		swr_on_60 = 3.6;
		update_swr_60(NULL);
	}
}

void run_40()
{
	flaa_status.swr_40_on = btn_run_40->value();
	if (flaa_status.swr_40_on) {
		swr_on_40 = 1.3;
		update_swr_40(NULL);
	}
}

void run_30()
{
	flaa_status.swr_30_on = btn_run_30->value();
}

void run_20()
{
	flaa_status.swr_20_on = btn_run_20->value();
}

void run_15()
{
	flaa_status.swr_15_on = btn_run_15->value();
}

void run_10()
{
	flaa_status.swr_10_on = btn_run_10->value();
}

void run_6()
{
	flaa_status.swr_6_on = btn_run_6->value();
}

void run_2()
{
	flaa_status.swr_2_on = btn_run_2->value();
}

void run_220()
{
	flaa_status.swr_220_on = btn_run_220->value();
}

void run_440()
{
	flaa_status.swr_440_on = btn_run_440->value();
}

void run_other()
{
	flaa_status.swr_other_on = btn_run_other->value();
}

void rf_on_off(int onoff)
{
	static char cmd[20];
	snprintf(
		cmd, sizeof(cmd), "FQ%d\n", 
		(int)(flaa_status.carrier_freq * 1e6 * (1 + flaa_status.freq_corr/1e6)));
	if (onoff) {
		write_command("ON\n");
//		MilliSleep(50);
		write_command(cmd);
	} else {
		write_command("OFF\n");
	}
}

void compute_velocity()
{
	if (!cable_values) cable_values = true;
}

void compute_length()
{
	if (!cable_values) cable_values = true;
}

Fl_Color color_select(Fl_Button *btn)
{
	Fl_Color clr = btn->color();
	unsigned char r, g, b;
	Fl::get_color(clr, r, g, b);
	if (fl_color_chooser("Font color", r, g, b, 1))
		clr = fl_rgb_color(r, g, b);
	return clr;
}
