// ----------------------------------------------------------------------------
// Copyright (C) 2014
//              David Freese, W1HKJ
//
// This file is part of flrig.
//
// flrig is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// flrig is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------

#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <FL/Fl_Preferences.H>
#include <FL/Fl_Progress.H>

#include <config.h>

#ifdef __linux__
#  include <dirent.h>
#  include <limits.h>
#  include <errno.h>
#  include <glob.h>
#endif
#ifdef __APPLE__
#  include <glob.h>
#endif
#ifndef __CYGWIN__
#  include <sys/stat.h>
#else
#  include <fcntl.h>
#endif

#include "Cartesian.H"

#include "status.h"
#include "util.h"
#include "flaa.h"
#include "debug.h"

status flaa_status = {
	50,			// int mainX;
	50,			// int mainY;
	600,		// int mainW;
	400,		// int mainH;
	"NONE",		// string serial_port;
	7,			// int baud 38400 default
	0,			// int units; 0=metric, 1=US
	50.0,		// double system_z
	10.1,		// double fmin in MHz
	10.2,		// double fmax in MHz
	50,			// int npts

	"",			// string port_ids;
	3.50,		// double	freq_80;
	5.72,		// double	freq_60;
	7.10,		// double	freq_40;
	10.155,		// double	freq_30;
	14.100,		// double	freq_20;
	21.100,		// double	freq_15;
	28.500,		// double	freq_10;
	52.1,		// double	freq_6;
	144.1,		// double	freq_2;
	223.0,		// double	freq_220;
	440.1,		// double	freq_440;
	0.0,		// double	freq_other;

	10.14,		// double	carrier_freq;
	0.0,		// double	freq_corr;

	4.8,		// double	cable_freq;
	0.66,		// double	velocity_factor;
	0.0,		// double	cable_length;

	0.25,		// double	wave;

	CA_NO_POINT,	// int	swr_point;
	FL_RED,			// int	swr_color;

	CA_NO_POINT,	// int	r_point;
	FL_GREEN,		// int	r_color;

	CA_NO_POINT,	// int	x_point;
	FL_BLUE,		// int	x_color;

	false,		// bool	swr_80_on;
	false,		// bool	swr_60_on;
	false,		// bool	swr_40_on;
	false,		// bool	swr_30_on;
	false,		// bool	swr_20_on;
	false,		// bool	swr_15_on;
	false,		// bool	swr_10_on;
	false,		// bool	swr_6_on;
	false,		// bool	swr_2_on;
	false,		// bool	swr_220_on;
	false,		// bool	swr_440_on;
	false		// bool	swr_other_on;

};

void status::saveLastState()
{

	Fl_Preferences spref(flaa_home_dir.c_str(), "w1hkj.com", PACKAGE_TARNAME);

	int mX = main_window->x();
	int mY = main_window->y();
	int mW = main_window->w();
	int mH = main_window->h();

	spref.set("version", PACKAGE_VERSION);

	spref.set("mainx", mX);
	spref.set("mainy", mY);
	spref.set("mainw", mW);
	spref.set("mainh", mH);

	spref.set("port", serial_port.c_str());
	spref.set("baudrate", baudrate);

	spref.set("units", units);
	spref.set("system_z", system_z);
	spref.set("fmin", fmin);
	spref.set("fmax", fmax);
	spref.set("npts", npts);

	spref.set("freq_80", freq_80);
	spref.set("freq_60", freq_60);
	spref.set("freq_40", freq_40);
	spref.set("freq_30", freq_30);
	spref.set("freq_20", freq_20);
	spref.set("freq_15", freq_15);
	spref.set("freq_10", freq_10);
	spref.set("freq_6", freq_6);
	spref.set("freq_2", freq_2);
	spref.set("freq_220", freq_220);
	spref.set("freq_440", freq_440);
	spref.set("freq_other", freq_other);

	spref.set("carrier_freq", carrier_freq);
	spref.set("freq_corr", freq_corr);
	spref.set("cable_freq", cable_freq);
	spref.set("velocity_factor", velocity_factor);
	spref.set("wave", wave);

	spref.set("swr_point", swr_point);
	spref.set("swr_color", swr_color);
	spref.set("r_point", r_point);
	spref.set("r_color", r_color);
	spref.set("x_point", x_point);
	spref.set("x_color", x_color);

}

void status::loadLastState()
{
	Fl_Preferences spref(flaa_home_dir.c_str(), "w1hkj.com", PACKAGE_TARNAME);

	if (spref.entryExists("version")) {

		int i;
		char defbuffer[500];

		spref.get("mainx", mainX, mainX);
		spref.get("mainy", mainY, mainY);
		spref.get("mainw", mainW, mainW);
		spref.get("mainh", mainH, mainH);

		spref.get("port", defbuffer, "NONE", 499);
		serial_port = defbuffer;
		if (serial_port.find("tty") == 0)
			serial_port.insert(0, "/dev/");
		spref.get("baudrate", baudrate, baudrate);

		spref.get("units", i, units);
		units = i;

		spref.get("system_z", system_z, system_z);

		spref.get("fmin", fmin, fmin);
		spref.get("fmax", fmax, fmax);
		spref.get("npts", npts, npts);

		spref.get("freq_80", freq_80, freq_80);
		spref.get("freq_60", freq_60, freq_60);
		spref.get("freq_40", freq_40, freq_40);
		spref.get("freq_30", freq_30, freq_30);
		spref.get("freq_20", freq_20, freq_20);
		spref.get("freq_15", freq_15, freq_15);
		spref.get("freq_10", freq_10, freq_10);
		spref.get("freq_6", freq_6, freq_6);
		spref.get("freq_2", freq_2, freq_2);
		spref.get("freq_220", freq_220, freq_220);
		spref.get("freq_440", freq_440, freq_440);
		spref.get("freq_other", freq_other, freq_other);

		spref.get("carrier_freq", carrier_freq, carrier_freq);
		spref.get("freq_corr", freq_corr, freq_corr);
		spref.get("cable_freq", cable_freq, cable_freq);
		spref.get("velocity_factor", velocity_factor, velocity_factor);
		spref.get("wave", wave, wave);

		spref.get("swr_point", swr_point, swr_point);
		spref.get("swr_color", swr_color, swr_color);
		spref.get("r_point", r_point, r_point);
		spref.get("r_color", r_color, r_color);
		spref.get("x_point", x_point, x_point);
		spref.get("x_color", x_color, x_color);

	}
	testCommPorts();

}

#ifdef __WIN32__
static bool open_serial(const char* dev)
{
	bool ret = false;
	HANDLE fd = CreateFile(dev, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
	if (fd != INVALID_HANDLE_VALUE) {
		CloseHandle(fd);
		ret = true;
	}
	return ret;
}

#endif //__WIN32__

void status::testCommPorts()
{
#ifndef PATH_MAX
#  define PATH_MAX 1024
#endif
#ifndef __WOE32__
	struct stat st;
#endif

#ifndef __APPLE__
	char ttyname[PATH_MAX + 1];
#endif

	const char* tty_fmt[] = {
#if defined(__linux__)
		"/dev/ttyS%u",
		"/dev/ttyUSB%u",
		"/dev/usb/ttyUSB%u",
		"/dev/ttyACM%u",
		"/dev/usb/ttyACM%u",
		"/opt/vttyS%u"
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
		"/dev/tty%2.2u"
#elif defined(__CYGWIN__)
		"/dev/ttyS%u"
#elif defined(__MINGW32__)
		"//./COM%u"
#elif defined(__APPLE__)
		"/dev/cu.*",
		"/dev/tty.*"
#endif
	};

#if defined(__WOE32__)
#  define TTY_MAX 255
#elif defined(__OpenBSD__) || defined(__NetBSD__)
#  define TTY_MAX 4
#else
#  define TTY_MAX 8
#endif

#ifdef __linux__
	glob_t gbuf;
	glob("/dev/serial/by-id/*", 0, NULL, &gbuf);
	for (size_t j = 0; j < gbuf.gl_pathc; j++) {
		if ( !(stat(gbuf.gl_pathv[j], &st) == 0 && S_ISCHR(st.st_mode)) ||
		     strstr(gbuf.gl_pathv[j], "modem") )
			continue;
		port_ids.append(gbuf.gl_pathv[j]).append("\n");
	}
	globfree(&gbuf);
#endif

	for (size_t i = 0; i < sizeof(tty_fmt)/sizeof(*tty_fmt); i++) {
#ifndef __APPLE__
		for (unsigned j = 0; j < TTY_MAX; j++) {
			snprintf(ttyname, sizeof(ttyname), tty_fmt[i], j);
#  ifndef __WOE32__
			if ( !(stat(ttyname, &st) == 0 && S_ISCHR(st.st_mode)) )
				continue;
#  else // __WOE32__
			if (!open_serial(ttyname))
				continue;
#    ifdef __CYGWIN__
			snprintf(ttyname, sizeof(ttyname), "COM%u", j+1);
#    else
			snprintf(ttyname, sizeof(ttyname), "COM%u", j);
#    endif
#  endif // __WOE32__

			port_ids.append(ttyname).append("\n");
		}
#else // __APPLE__
		glob_t gbuf;
		glob(tty_fmt[i], 0, NULL, &gbuf);
		for (size_t j = 0; j < gbuf.gl_pathc; j++) {
			int ret1 = !stat(gbuf.gl_pathv[j], &st);
			int ret2 = S_ISCHR(st.st_mode);
			if ( (ret1 && ret2 ) || strstr(gbuf.gl_pathv[j], "modem") )
				port_ids.append(gbuf.gl_pathv[j]).append("\n");
			else
				continue;
		}
		globfree(&gbuf);
#endif // __APPLE__
	}

#if HAVE_UHROUTER
	if (stat(UHROUTER_FIFO_PREFIX "Read", &st) != -1 && S_ISFIFO(st.st_mode) &&
	    stat(UHROUTER_FIFO_PREFIX "Write", &st) != -1 && S_ISFIFO(st.st_mode))
	    port_ids.append(UHROUTER_FIFO_PREFIX).append("\n");
#endif // HAVE_UHROUTER

}

